#  Copyright (c) 2020 Boulder Real Time Technologies, Inc.
#
#  This software module is wholly owned by Boulder Real Time
#  Technologies, Inc. Any use of this software module without
#  express written permission from Boulder Real Time Technologies,
#  Inc. is prohibited.


import os
import random
import logging
from datetime import datetime

from antelope.datascope import dbopen, closing, freeing, Dbptr, DbsubsetError
from webservice_fdsn.app.core.constants.geographic_constraints import AREA_RECTANGLE_MIN_LATITUDE,\
    AREA_RECTANGLE_MAX_LATITUDE, AREA_RECTANGLE_MIN_LONGITUDE, AREA_RECTANGLE_MAX_LONGITUDE
from webservice_fdsn.app.core.constants.fdsnws_event import MIN_DEPTH, MAX_DEPTH, MIN_MAGNITUDE, MAX_MAGNITUDE
from webservice_fdsn.app.core.event_exporter.export_events_helpers import exceptions
from webservice_fdsn.app.core.constants.time_constraints import SIMPLE_TIME_START_TIME, SIMPLE_TIME_END_TIME

log = logging.getLogger(__name__)


def apply_user_filters(dbview: Dbptr, **kwargs) -> Dbptr:
    """Apply user defined constraints to the passed in db view

    :param antelope.datascope.Dbptr dbview:
    :param dict kwargs: filter values
    :rtype: Dbptr
    :return: reference to filtered view
    """
    _subset_expr = "lat >= {minlat} && lat <= {maxlat} && " \
                   "lon >= {minlon} && lon <= {maxlon}" \
        .format(minlat=kwargs[AREA_RECTANGLE_MIN_LATITUDE],
                maxlat=kwargs[AREA_RECTANGLE_MAX_LATITUDE],
                minlon=kwargs[AREA_RECTANGLE_MIN_LONGITUDE],
                maxlon=kwargs[AREA_RECTANGLE_MAX_LONGITUDE])
    log.debug(f'Using lat/lon subset expr as: {_subset_expr}')
    try:
        sb1 = dbview.subset(_subset_expr)
    except DbsubsetError:
        _msg = 'No entries left after applying lat/lon constraints'
        log.warning(_msg)
        raise exceptions.NoEventDataFound(_msg)

    if sb1.record_count <= 0:
        raise exceptions.NoEventDataFound('No event data found after applying lat/lon filters. Your values: '
                                          'Minlat: {}, Maxlat: {}, Minlon: {}, Maxlon: {}'
                                          .format(kwargs[AREA_RECTANGLE_MIN_LATITUDE],
                                                  kwargs[AREA_RECTANGLE_MAX_LATITUDE],
                                                  kwargs[AREA_RECTANGLE_MIN_LONGITUDE],
                                                  kwargs[AREA_RECTANGLE_MAX_LONGITUDE]))
    log.debug(f'Entries after lat/lon filtering: {sb1.record_count}')

    _subset_expr = "depth >= {mindepth} && depth <= {maxdepth}".format(mindepth=kwargs[MIN_DEPTH],
                                                                       maxdepth=kwargs[MAX_DEPTH])
    log.debug(f'Using min/max depth subset expr as: {_subset_expr}')
    try:
        sb2 = sb1.subset(_subset_expr)
    except DbsubsetError:
        _msg = 'No entries left after applying depth constraints'
        log.warning(_msg)
        raise exceptions.NoEventDataFound(_msg)

    if sb2.record_count <= 0:
        raise exceptions.NoEventDataFound('No event data found after applying depth filters. Your values: '
                                          '{}: {}, {}: {}'.format(MIN_DEPTH, kwargs[MIN_DEPTH],
                                                                  MAX_DEPTH, kwargs[MAX_DEPTH]))
    log.debug(f'Entries after depth filtering: {sb2.record_count}')

    # Magnitude filtering not working as expected
    _subset_expr = "magnitude >= {minmag} && magnitude <= {maxmag}".format(minmag=kwargs[MIN_MAGNITUDE],
                                                                           maxmag=kwargs[MAX_MAGNITUDE])
    log.debug(f'Using min/max magnitude subset expr as: {_subset_expr}')
    try:
        sb3 = sb2.subset(_subset_expr)
    except DbsubsetError:
        _msg = 'No entries left after applying magnitude constraints'
        log.warning(_msg)
        raise exceptions.NoEventDataFound(_msg)

    if sb3.record_count <= 0:
        raise exceptions.NoEventDataFound('No event data found after applying magnitude filters. Your values: '
                                          'Minmag: {}, Maxmag: {}'.format(kwargs[MIN_MAGNITUDE],
                                                                          kwargs[MAX_MAGNITUDE]))
    log.debug(f'Entries after magnitude filtering: {sb3.record_count}')
    return sb3


def apply_time_window(table: Dbptr,
                      start_time: datetime.timestamp,
                      end_time: datetime.timestamp) -> Dbptr:
    """
    Apply time window filter to db
    :param antelope.datascope.Dbptr table:
    :param datetime.timestamp start_time:
    :param datetime.timestamp end_time:
    :rtype Dbptr
    :raises exceptions.NoEventDataFound: Raised if time constaints filter all data out
    :return: reference to filtered view
    """
    log.debug(f'Filtering event data using start/end times of: {start_time}/{end_time}. '
              f'Number of records before filtering: {table.record_count}')
    try:
        _subset = f'origin.time >= "{start_time}" && origin.time <= "{end_time}"'
        log.debug(f'Using subset expr: {_subset}')
        result_dbptr = table.subset(_subset)
    except DbsubsetError:
        _msg = 'No results found after applying time window constraints'
        log.warning(_msg)
        raise exceptions.NoEventDataFound(_msg)
    log.debug(f'Number of records after time window filtering: {result_dbptr.record_count}')
    return result_dbptr


def get_event_data(db_path: str, scratch_dir: str, **kwargs) -> str:
    """Returns requested event data view

    :param str db_path: path to database
    :param str scratch_dir: scratch dir path where event data file will be written
    :param kwargs:
    :rtype str
    :return: path to event data file
    """
    # Important to set noshare=1 to handle concurrent web requests
    with closing(dbopen(db_path, 'r', noshare=1)) as db:
        dbtable = db.lookup(table='event')
        
        # First, apply the time constraints before joining to get other data
        # Need to filter by origin.time, join with origin table before applying
        # time window constraints.
        with freeing(dbtable.join('origin', outer=True)) as origin_join_view:
            time_filtered_db_table = apply_time_window(origin_join_view,
                                                       kwargs[SIMPLE_TIME_START_TIME],
                                                       kwargs[SIMPLE_TIME_END_TIME])

            with freeing(time_filtered_db_table.join('netmag', outer=True)) as dbview:
                log.debug(f'Record count after netmag: {dbview.record_count}')
                if dbview.record_count <= 0:
                    raise exceptions.NoEventDataFound('No events found after netmag join')
                subset = apply_user_filters(dbview, **kwargs)
                log.debug(f'Record count after subset: {subset.record_count}')
                data_path = os.path.join(scratch_dir, f'dbe_{random.randint(1, 10000)}')
                subset.unjoin(data_path)
                log.debug(f'Writing event data to file {data_path}')
                return data_path


if __name__ == "__main__":
    get_event_data("/Users/rohan_work/rtdemo/db/gsn",
                   "/tmp",
                   minlatitude=33, maxlatitude=34,
                   minlongitude=-117.34, maxlongitude=-115.34,
                   mindepth=0, maxdepth=200,
                   minmagnitude=1.0, maxmagnitude=10.0)
