#   Copyright (c) 2020 Boulder Real Time Technologies, Inc.
#
#   This software module is wholly owned by Boulder Real Time
#   Technologies, Inc. Any use of this software module without
#   express written permission from Boulder Real Time Technologies,
#   Inc. is prohibited.
#
#   Author: Ambli LLC (ambli.ll.company@gmail.com)

"""Module to run XML postprocessing operations on XML data returned by db2stationxml"""

import os
import re
from dateutil import parser as dt_parser
from datetime import datetime
from lxml import etree
import xml.etree.ElementTree as ET
import logging

# Looking for namespace in xml tag: "{http://something.com}Tag"
namespace_regex = re.compile('^{([\w:\-\/\.]+)}([\w]+)$')

log = logging.getLogger(__name__)


def _filter_stations_by_time_window(root: ET.Element, ns: dict,
                                    starttime: datetime,
                                    endtime: str=None) -> bool:
    """Return stations which fall inside the timewindow constraints

    :param ET.Element root: XML root element
    :param datetime starttime: time window start
    :param string endtime: time window end
    :rtype bool
    :return: True if stations matched the filters
    """

    for network in root.findall('ns:Network', ns):
        stations = network.findall('ns:Station', ns)
        log.debug(f'Found {len(stations)} stations in network {network.attrib.get("code")} '
                  f'before time filtering')
        if not stations:
            # If nothing came in, nothing to do
            continue

        for stn in stations:
            st_start = dt_parser.parse(stn.attrib.get('startDate'))
            st_end = dt_parser.parse(stn.attrib.get('endDate', '2599-12-31 23:59:59.99900'))
            if not (st_start and st_end):
                # If station missing start/end date, throw it away
                # TODO: Maybe missing end_date is OK? Check with Kent
                log.debug(f'Discarding station {stn.attrib.get("code")} due to missing start/end times')
                network.remove(stn)

            if endtime:
                if (st_start < starttime) or (st_end > endtime):
                    log.debug(f'Removing station {stn.attrib.get("code")} as start/end times don\'t match up')
                    network.remove(stn)
            else:
                if st_start < starttime:
                    log.debug(f'Removing station {stn.attrib.get("code")} as station start time is before '
                              f'requested starttime')
                    network.remove(stn)

        stations = network.findall('ns:Station', ns)
        if not stations:
            log.debug(f'No stations left in network {network.attrib.get("code")} after time filtering, '
                      f'removing network from results')
            root.remove(network)
        else:
            log.debug(f'Left with {len(stations)} stations in network {network.attrib.get("code")} after '
                      f'time filtering')

    # Filtering done, is anything left?
    for network in root.findall('ns:Network', ns):
        stations = network.findall('ns:Station', ns)
        if stations:
            return True
    return False


def _filter_stations_by_coordinates(root: ET.Element, ns: dict,
                                    minlat: float, maxlat: float,
                                    minlon: float, maxlon: float) -> bool:
    """Return stations which fall inside the geographic constraints

    :param ET.Element root:
    :param dict ns: xml namespace info
    :param float minlat: geographic constraint
    :param float maxlat: geographic constraint
    :param float minlon: geographic constraint
    :param float maxlon: geographic constraint
    :rtype bool
    :return: True if stations matched the filters
    """

    for network in root.findall('ns:Network', ns):
        stations = network.findall('ns:Station', ns)
        log.debug(f'Found {len(stations)} stations in network {network.attrib.get("code")} '
                  f'before coordinate filtering')
        if not stations:
            # If nothing came in, nothing to do
            continue
        for stn in stations:
            lat = stn.find('ns:Latitude', ns)
            lon = stn.find('ns:Longitude', ns)

            if not (lat.text and lon.text):
                # If station missing lat/lon info, throw it away
                log.debug(f'Discarding station {stn.attrib.get("code")} due to missing lat/lon info')
                network.remove(stn)

            if not((minlat <= float(lat.text) <= maxlat) and (minlon <= float(lon.text) <= maxlon)):
                log.debug(f'Removing station {stn.attrib.get("code")}')
                network.remove(stn)

        stations = network.findall('ns:Station', ns)
        if not stations:
            log.debug(f'No stations left in network {network.attrib.get("code")} after time filtering, '
                      f'removing network from results')
            root.remove(network)
        else:
            log.debug(f'Left with {len(stations)} stations in network {network.attrib.get("code")} '
                      f'after lat/lon filtering')

    # Filtering done, is anything left?
    for network in root.findall('ns:Network', ns):
        stations = network.findall('ns:Station', ns)
        if stations:
            return True
    return False


def get_networks(root, ns: dict) -> list:
    """Return network codes from xml root
    :rtype list
    :return: list of network codes
    """
    return [network.attrib.get('code') for network in root.findall('ns:Network', ns)]


def get_stations(root, ns: dict) -> list:
    """Return station elements from xml root

    :param root:
    :param ns:
    :rtype list
    :return: list of station elements
    """
    stations = []
    for network in root.findall('ns:Network', ns):
        stations.extend([station.attrib.get('code') for station in network.findall('ns:Station', ns)])
    return stations


def validate_stationxml(xml_path: str) -> bool:
    """Run simple XML validation using schema file
    :param xml_path:
    :param xsd_path:
    :return:
    """
    cur_path = os.path.dirname(__file__)
    stationxml_xsd = os.path.join(cur_path, 'schema', 'fdsn-station.xsd')
    xmlschema_doc = etree.parse(stationxml_xsd)
    xmlschema = etree.XMLSchema(xmlschema_doc)

    xml_doc = etree.parse(xml_path)
    return xmlschema.validate(xml_doc)


def get_xml_root(xmlpath: str) -> ET.Element:
    """Return rool element from xml document

    :param str xmlpath: path to xml document
    :rtype ET.Element
    :return: root element
    """
    # inventory is really the collection of all networks, stations and channels included in this StationXML file
    inventory = ET.parse(xmlpath)
    return inventory.getroot()


def get_xml_namespace(root: ET.Element) -> dict:
    """Get namespace dictionary

    :param ET.Element root:
    :rtype dict
    :return:
    """
    ns_info = namespace_regex.search(root.tag)
    if ns_info:
        return {'ns': ns_info.groups()[0]}


def db2stationxml_filter(xmlpath: str, xmloutputpath: str,
                         minlatitude: float, maxlatitude: float,
                         minlongitude: float, maxlongitude: float,
                         starttime: datetime, endtime: datetime=None) -> bool:
    """db2stationxml does not support filtering queries based on lat/lon and time windows which
    the FDSN spec requires. Post-process XML data returned by db2stationxml which filters by
    sta/chan/net/loc and apply lat/lon and time window constraints
    :param string xmlpath: path to xml file
    :param string xmloutputpath: path to write output XML to
    :param float minlatitude: geographic constraint
    :param float maxlatitude: geographic constraint
    :param float minlongitude: geographic constraint
    :param float maxlongitude: geographic constraint
    :param datetime.datetime starttime: time window start
    :param datetime.datetime endtime: time window end
    :rtype bool
    :return: True if any stations matched user query params
    """
    # Obspy datetime objects are timezone naive, remove them from user input as well
    starttime = starttime.replace(tzinfo=None)
    if endtime:
        endtime = endtime.replace(tzinfo=None)

    root = get_xml_root(xmlpath)

    # Get namespace info
    ns = get_xml_namespace(root)

    _data_match = _filter_stations_by_coordinates(root, ns, minlatitude, maxlatitude,
                                                  minlongitude, maxlongitude)
    if _data_match:
        _data_match &= _filter_stations_by_time_window(root, ns, starttime, endtime)
    else:
        log.debug('No stations left after coordinate filtering, skipping time filtering...')

    if _data_match:
        with open(xmloutputpath, 'w+') as f:
            f.write(ET.tostring(root).decode('utf-8'))

    return _data_match


if __name__ == '__main__':
    db2stationxml_filter('/Users/rohan_work/Downloads/response_1579481983653.xml',
                         '/Users/rohan_work/Downloads/xmldump', -90, 90, -146, 180,
                         '2000-01-01T00:00:00', None)
