import os
import sys

sys.path.append(os.environ['ANTELOPE'] + "/data/python")

from antelope import stock
from webservice_fdsn.app.utils import exceptions
from webservice_fdsn.app.core.constants import common


def get_pf_options():
    """Returns values from pf file in dict format.

    :rtype dict
    :return: dictionary with pf options
    """
    # Add support to source pfname from env-var to ease testing
    pfname = os.environ.get(common.ENV_VAR_PF_PATH, 'webservice_fdsn')
    try:
        pf_handle = stock.pfread(pfname)
    except stock.PfReadError as ex:
        raise exceptions.InvalidPfFile('Supplied pf file does not exist or is improperly '
                                       f'formatted, error: {ex}')
    return pf_handle.pf2dict()


def get_project_logging_path():
    """Return where the application should dump its logs.

    :rtype: str
    :return:
    """
    return get_pf_options().get('app_log_dir', '/tmp/')


def get_project_scratch_dir():
    """Return user defined scratch directory path.

    :rtype str
    :return: absolute path to scratch/working directory where service can store intermediate files as part of
    servicing user requests
    """
    return get_pf_options().get('app_scratch_dir', '/tmp/')


def get_default_service_network():
    """If a service is tied to a specific network, return it so queries for other networks can be rejected.

    :rtype str/None
    :return:
    """
    return get_pf_options().get('app_default_network')


def get_project_database_path():
    """Return path to antelope db as defined in the parameter file.

    :rtype str
    :return: path to db
    """
    pf = get_pf_options()
    if 'dbname' not in pf:
        raise exceptions.RequiredPfOptionMissing('pf file must specify path to Antelope DB via dbname param')

    try:
        os.stat(pf['dbname'])
    except FileNotFoundError:
        raise exceptions.InvalidPfValue('DB path invalid. Path does not exist: {}'.format(pf['dbname']))

    return pf['dbname']


def get_cluster_name_for_api(fdsn_api_group):
    """Return name of dbcluster entry corresponding to api name

    If user is using dbcluster, return name of cluster in dbcluster file
    corresponding to event, dataselect or station FDSN api group
    :param str fdsn_api_group: station, event or dataselect
    :rtype str
    :return: name of cluster
    """
    key_name = f'clustername_{fdsn_api_group}'
    pf = get_pf_options()
    if key_name not in pf:
        raise exceptions.RequiredPfOptionMissing('Looks like you are using dbcluster. Your pf file must specify cluster name '
                                                 f'for {fdsn_api_group} API, format: "clustername_<event/station/dataselect>"')

    return pf[key_name]


def get_app_configuration_options():
    """Return pf values for configuration the web interface

    :rtype: dict
    """
    pf = get_pf_options()
    app_config_pf_options = {(k, v) for (k, v) in pf.items() if k.startswith('app_config')}
    return app_config_pf_options
