#   Copyright (c) 2020 Boulder Real Time Technologies, Inc.
#
#   This software module is wholly owned by Boulder Real Time
#   Technologies, Inc. Any use of this software module without
#   express written permission from Boulder Real Time Technologies,
#   Inc. is prohibited.
#
#   Author: Ambli LLC (ambli.ll.company@gmail.com)

import connexion
import logging
import six
import os
import glob
from http import HTTPStatus
from flask import make_response
from dateutil import parser
from datetime import datetime, timezone

from webservice_fdsn.app.utils import pf_helper
from webservice_fdsn.app.core import constants
from webservice_fdsn.app.utils import exceptions as apiutil_exceptions
from webservice_fdsn.app.utils import utilities
from webservice_fdsn.app.core.event_helpers import get_event_data
from webservice_fdsn.app.core.event_exporter import event2qml
from webservice_fdsn.app.core.event_exporter.export_events_helpers import exceptions

log = logging.getLogger(__name__)


def get_event(mindepth, maxdepth, minmagnitude, maxmagnitude, orderby, starttime,
              minlatitude, maxlatitude, minlongitude, maxlongitude,
              includeallorigins=None, eventid=None, limit=None, offset=None, endtime=None, nodata=None):  # noqa: E501
    """Controller handling all API requests to get event data

     # noqa: E501

    :param mindepth: Limit to events with depth (in Kilometers) more than the specified minimum.
    :type mindepth: float
    :param maxdepth: Limit to events with depth (in Kilometers) less than the specified maximum.
    :type maxdepth: float
    :param minmagnitude: Limit to events with a magnitude larger than the specified minimum.
    :type minmagnitude: float
    :param maxmagnitude: Limit to events with a magnitude smaller than the specified maximum.
    :type maxmagnitude: float
    :param orderby: Order the result by time or magnitude with the following possibilities: time: order by origin descending time time-asc: order by origin ascending time magnitude: order by descending magnitude magnitude-asc: order by ascending magnitude
    :type orderby: str
    :param starttime: Limit to events on or after the specified start time (UTC)
    :type starttime: str
    :param minlatitude: Limit to events with a latitude larger than or equal to the specified minimum.
    :type minlatitude: float
    :param maxlatitude: Limit to events with a latitude smaller than or equal to the specified maximum.
    :type maxlatitude: float
    :param minlongitude: Limit to events with a longitude larger than or equal to the specified minimum.
    :type minlongitude: float
    :param maxlongitude: Limit to events with a longitude smaller than or equal to the specified maximum.
    :type maxlongitude: float
    :param includeallorigins: Specify if all origins for the event should be included, default is data center dependent but is suggested to be the preferred origin only.
    :type includeallorigins: bool
    :param eventid: Select a specific event by ID; event identifiers are data center specific.
    :type eventid: int
    :param limit: Limit the results to the specified number of events.
    :type limit: int
    :param offset: Return results starting at the event count specified, starting at 1.
    :type offset: int
    :param endtime: Limit to events on or before the specified end time (UTC)
    :type endtime: str
    :param nodata: Select status code for no data; either 204 (default) or 404.
    :type nodata: str

    :rtype: HTTPResponse
    """
    log.info(f'Querying event data using params: mindepth: {mindepth}, maxdepth: {maxdepth}, minmagnitude: {minmagnitude}, '
             f'maxmagnitude: {maxmagnitude}, orderby: {orderby}, starttime: {starttime}, eventid: {eventid}, '
             f'minlatitude: {minlatitude}, maxlatitude: {maxlatitude}, includeallorigins: {includeallorigins}, '
             f'minlongitude: {minlongitude}, maxlongitude: {maxlongitude}, endtime: {endtime} '
             f'limit: {limit}, offset: {offset}')

    # Get scratch dir path for intermediate files
    scratch_dir = pf_helper.get_project_scratch_dir()

    # Convert times to timestamp format
    starttime = parser.parse(starttime)
    starttime = starttime.replace(tzinfo=None)
    if endtime:
        endtime = parser.parse(endtime)
        endtime = endtime.replace(tzinfo=None)
    else:
        endtime = datetime.utcnow()
    if includeallorigins:
        if includeallorigins == 'true':
            includeallorigins = True
        else:
            includeallorigins = False

    # Validate limit and offset
    if limit is not None:
        limit = int(limit)
        # If user passes limit of 0, short-circuit here and return 'nodata' without doing any work
        if limit == 0:
            return make_response('Limit set to 0, nothing to return', nodata)
        if limit < 0:
            return make_response('Limit has to be a positive integer', HTTPStatus.BAD_REQUEST)
    if offset is not None:
        offset = int(offset)
        if offset < 0:
            return make_response('Offset has to be a positive integer', HTTPStatus.BAD_REQUEST)

    # Get database file/s on disk
    try:
        db_paths = utilities.get_db_files(constants.service_interfaces_constants.EVENT,
                                          scratch_dir, starttime, endtime)
    except (apiutil_exceptions.RequiredPfOptionMissing,
            apiutil_exceptions.InvalidPfValue) as ex:
        log.error(f"Service configuration error: {ex}")
        resp = make_response(f'{ex}', HTTPStatus.INTERNAL_SERVER_ERROR)
        return resp

    if not db_paths:
        resp = make_response("", nodata)

    log.debug(f'Searching event data in {len(db_paths)} files')
    # Event data could be in a single file for css3.0 dbs, or spread across multiple files if
    # data is organized using dbcluster scheme. Iterate over all databases, and combine all quakeml
    # data before returning to user

    resp = make_response("", nodata)
    qmls = []

    for db_path in db_paths:
        log.debug(f'>>> Looping to fetch event data from {db_path}')
        view_db = None
        try:
            # Get filtered view of events by applying user defined constraints.
            view_db = get_event_data.get_event_data(db_path, scratch_dir, starttime=starttime, endtime=endtime,
                                                    minlatitude=minlatitude, minlongitude=minlongitude,
                                                    maxlatitude=maxlatitude, maxlongitude=maxlongitude,
                                                    mindepth=mindepth, maxdepth=maxdepth,
                                                    minmagnitude=minmagnitude, maxmagnitude=maxmagnitude)
        except exceptions.NoEventDataFound as ex:
            log.debug(f'No event data found in {db_path} for request parameters')
            # No data found in this db, continue looking in other dbs if available
            continue
        except Exception as ex:
            log.error(f'Internal server error: {ex}')
            resp = make_response(format(ex), HTTPStatus.INTERNAL_SERVER_ERROR)
        else:
            if view_db:
                try:
                    # Use generated view to extract event information in quakeml format.
                    # TODO: Get name of event2qml pf file to use from event2qml module
                    quakeml = event2qml.event2qml('event2qml.pf', view_db, event_id=eventid,
                                                  includeallorigins=includeallorigins,
                                                  order_by=orderby)
                    
                except exceptions.EventNotFound:
                    # EventNotFound raised when a particular event_id is not found
                    # Continue looking in other databases in this case
                    log.debug(f'No event with ID {eventid} found, continuing search...')
                    #resp = make_response(f'No event with ID {eventid} found', nodata)
                except exceptions.NoEventDataFound as ex:
                    # No event data found in this view, continue searching others views
                    # if available
                    log.debug(f'No event data found in {db_path} for user request')
                else:
                    # Collect all quakemls, combine them in the end
                    qmls.append(quakeml)

                    # Clean up intermediate view files
                    db_files = glob.glob(f'{view_db}*')
                    for f in db_files:
                        log.debug(f'Cleaning up intermediate file {f}')
                        os.remove(f)

                    # if found event ID of interest, break out early
                    if quakeml and eventid:
                        log.debug('Found requested event ID.')
                        break

            else:
                log.warning(f'No event data found in {db_path} for user request')
    if qmls:
        # combine into single quakeml object. Handle the limit and offset params here
        try:
            resp = make_response(utilities.combine_quakemls(qmls, limit=limit, offset=offset), HTTPStatus.OK)
        except exceptions.NoEventDataFound as ex:
            resp = make_response(f'{ex}', nodata)
    else:
        if eventid:
            log.debug(f'Did not find requested event ID {eventid} ')
        resp = make_response("", nodata)

    return resp


def get_event_catalogs():  # noqa: E501
    """Returns a canned response

    :rtype: None
    """
    canned_data = """<Catalogs>
        <Catalog>EMSC</Catalog>
        <Catalog>GCMT</Catalog>
        <Catalog>NEIC PDE</Catalog>
        <Catalog>ISC</Catalog>
    </Catalogs>"""
    resp = make_response(canned_data, HTTPStatus.OK)
    return resp


def get_event_contributers():  # noqa: E501
    """Returns a canned response

    :rtype: Response
    """
    canned_data = """<Contributors>
        <Contributor>NEIC</Contributor>
        <Contributor>EMSC</Contributor>
        <Contributor>ISC</Contributor>
    </Contributors>"""
    resp = make_response(canned_data, HTTPStatus.OK)
    return resp


def get_event_version():
    """get_event_version

    :rtype: Response
    """
    resp = make_response("1.0.4", HTTPStatus.OK)
    return resp


def get_event_wadl():
    """get_event_wadl

    :rtype: Response
    """
    wadl_path = os.path.join(os.path.dirname(__file__), os.path.pardir, "wadl", "event.wadl")
    with open(wadl_path, "r") as f:
        resp = make_response("".join(f.readlines()), HTTPStatus.OK)

        return resp
