import pytest
import urllib3
from obspy import UTCDateTime
from obspy.clients.fdsn import Client
from obspy.clients.fdsn.header import FDSNException, FDSNNoDataException

from tests.deployment_validation import util


ip_op_400_matrix = [{},
                    {'starttime': '2010-01-01 00:00:00'},
                    {'starttime': '2010-01-01 00:00:00', 'network': 'AZ'},
                    {'starttime': '2010-01-01 00:00:00', 'network': 'AZ', 'station': 'B.*'},
                    {'starttime': '2010-01-01 00:00:00', 'network': 'AZ', 'station': 'B.*', 'channel': '.*'},
                    {'starttime': '2010-01-01 00:00:00', 'network': 'AZ', 'station': 'B.*', 'channel': '.*',
                     'location': '.*'},
                    {'starttime': '2010-01-01 00:00:00', 'network': 'AZ', 'station': 'B.*', 'channel': '.*',
                     'location': '.*', 'minlatitude': -90},
                    {'starttime': '2010-01-01 00:00:00', 'network': 'AZ', 'station': 'B.*', 'channel': '.*',
                     'location': '.*', 'minlatitude': -90, 'maxlatitude': 90},
                    {'starttime': '2010-01-01 00:00:00', 'network': 'AZ', 'station': 'B.*', 'channel': '.*',
                     'location': '.*', 'minlatitude': -90, 'maxlatitude': 90, 'minlongitude': -180}
                    ]


@pytest.mark.parametrize("query_params", ip_op_400_matrix)
def test_event_query_response_codes(server, query_params):
    api = util.load_server(server, None)
    url = urllib3.util.parse_url(api)
    host = f"{url.scheme}://{url.host}"
    if url.port:
        host = f"{host}:{url.port}"
    client = Client(host)
    with pytest.raises(FDSNException):
        client.get_stations(**query_params)


query_all = [{'starttime': '2010-01-01 00:00:00', 'network': '.*', 'station': '.*', 'channel': '.*',
              'location': '.*', 'minlatitude': -90, 'maxlatitude': 90,
              'minlongitude': -180, 'maxlongitude': 180, 'nodata': '204', 'level': 'station'}]


@pytest.mark.parametrize("query_all", query_all)
def test_no_data(server, query_all):
    api = util.load_server(server, None)
    query_all['network'] = 'AZZZZZZ'
    url = urllib3.util.parse_url(api)
    host = f"{url.scheme}://{url.host}"
    if url.port:
        host = f"{host}:{url.port}"
    client = Client(host)
    # Should raise FDSNNoDataException, but obspy debug says this:
    # "UserWarning: The parameter 'nodata' is ignored because it is not useful within ObsPy"
    with pytest.raises(FDSNException):
        client.get_stations(**query_all)


@pytest.mark.parametrize("query_all", query_all)
def test_network_filter(server, query_all):
    api = util.load_server(server, None)
    query_all['network'] = 'AZ'
    url = urllib3.util.parse_url(api)
    host = f"{url.scheme}://{url.host}"
    if url.port:
        host = f"{host}:{url.port}"
    client = Client(host)
    stations = client.get_stations(**query_all)

    assert 'AZ' in [nw.code for nw in stations.networks]


@pytest.mark.parametrize("query_all", query_all)
def test_station_filter(server, query_all):
    api = util.load_server(server, None)
    query_all['network'] = 'AZ'
    query_all['station'] = 'B.*'
    url = urllib3.util.parse_url(api)
    host = f"{url.scheme}://{url.host}"
    if url.port:
        host = f"{host}:{url.port}"
    client = Client(host)
    stations = client.get_stations(**query_all)

    for st in stations.networks[0].stations:
        assert st.code.startswith('B')


@pytest.mark.parametrize("query_all", query_all)
def test_latitude_filter(server, query_all):
    api = util.load_server(server, None)
    query_all['minlatitude'] = '33'
    query_all['maxlatitude'] = '34'
    url = urllib3.util.parse_url(api)
    host = f"{url.scheme}://{url.host}"
    if url.port:
        host = f"{host}:{url.port}"
    client = Client(host)
    stations = client.get_stations(**query_all)
    stations = stations.networks[0].stations
    for st in stations:
        assert 33 <= st.latitude <= 34


@pytest.mark.parametrize("query_all", query_all)
def test_longitude_filter(server, query_all):
    api = util.load_server(server, None)
    query_all['minlongitude'] = '-117'
    query_all['maxlongitude'] = '-116'
    url = urllib3.util.parse_url(api)
    host = f"{url.scheme}://{url.host}"
    if url.port:
        host = f"{host}:{url.port}"
    client = Client(host)
    stations = client.get_stations(**query_all)
    stations = stations.networks[0].stations
    for st in stations:
        assert -117 <= st.longitude <= -116


@pytest.mark.parametrize("query_all", query_all)
def test_starttime_filter(server, query_all):
    api = util.load_server(server, None)
    query_all['starttime'] = '2011-08-11 00:00:00'
    url = urllib3.util.parse_url(api)
    host = f"{url.scheme}://{url.host}"
    if url.port:
        host = f"{host}:{url.port}"
    client = Client(host)
    stations = client.get_stations(**query_all)
    stations = stations.networks[0].stations
    for st in stations:
        assert UTCDateTime(year=2011, month=8, day=11) <= st.start_date


@pytest.mark.parametrize("query_all", query_all)
def test_endtime_filter(server, query_all):
    api = util.load_server(server, None)
    query_all['endtime'] = '2016-12-31 00:00:00'
    url = urllib3.util.parse_url(api)
    host = f"{url.scheme}://{url.host}"
    if url.port:
        host = f"{host}:{url.port}"
    client = Client(host)
    stations = client.get_stations(**query_all)
    stations = stations.networks[0].stations
    for st in stations:
        assert st.end_date <= UTCDateTime(year=2016, month=12, day=31)


@pytest.mark.parametrize("query_all", query_all)
def test_level(server, query_all):
    api = util.load_server(server, None)
    query_all['level'] = 'station'
    query_all['network'] = 'AZ.*'
    query_all['station'] = 'B.*'
    url = urllib3.util.parse_url(api)
    host = f"{url.scheme}://{url.host}"
    if url.port:
        host = f"{host}:{url.port}"
    client = Client(host)
    stations = client.get_stations(**query_all)
    assert len(stations.networks[0].stations[0].channels) == 0

    query_all['level'] = 'channel'
    stations = client.get_stations(**query_all)
    assert len(stations.networks[0].stations[0].channels) > 0
